import os
import pwd
from builtins import bool
import logging
from pypka import __version__ as pypka_version
from pypka.config import ParametersDict as ParametersDictPypKa
from pypkamd.misc import remove_comments, read_ff_dict_section


def get_username():
    return pwd.getpwuid(os.getuid())[0]


class Config:
    @classmethod
    def storeParams(cls, mdp: str):
        pypka_version_parts = pypka_version.split(".")
        minimum_parts = "2.3.0".split(".")
        needs_upgrade = sum(
            [
                0 if int(part) >= int(minimum_parts[i]) else 1
                for i, part in enumerate(pypka_version_parts)
            ]
        )
        if needs_upgrade:
            raise Exception(
                "Please update your PypKa installation. \nCurrent version: {}\nRequired version: >= 2.3.0".format(
                    pypka_version
                )
            )

        cls.md_configs = MDConfig(mdp)
        cls.md_configs.read_input_mdp()


class ParametersDict(ParametersDictPypKa):
    input_conversion = {}
    input_conversion["dt"] = "TimeStep"
    input_conversion["tinit"] = "InitTime"
    input_conversion["ref_t"] = "temp"
    input_conversion["nsteps"] = "EffectiveSteps"


class MDConfig(ParametersDict):
    def __init__(self, fmdp: str):

        self.TimeStep = None
        self.InitTime = None
        self.EffectiveSteps = None
        self.RelaxSteps = 100

        self.nCycles = 0

        self.InitCycle = 0

        self.temp = None

        self.tmpDIR = "./tmp"
        self.script_dir = os.path.dirname(os.path.abspath(__file__))

        self.ff_options = {"CHARMM36mpH": "CHARMM36m", "G54A7pH": "G54A7"}
        self.ffID = None

        self.USER = get_username()
        self.HOST = os.uname()[1]

        self.fixboxDIR = "{0}/fixbox/".format(self.script_dir)

        self.effective_name = "effective"
        self.relax_name = "relax"

        self.MDP = f"{self.effective_name}.mdp"
        self.MDP_relax = f"{self.relax_name}.mdp"
        self.GRO = f"{self.effective_name}.gro"
        self.GRO_relax = f"{self.relax_name}.gro"

        self.sysname = None

        self.TOP = None
        self.NDX = None

        self.GROin = None
        self.TOPin = None
        self.MDPin = fmdp
        self.DATin = ""
        self.NDXin = ""

        self.LOG_fname = "LOG_CpHMD"
        self.LOG = open(self.LOG_fname, "a")

        self.GroDIR = None

        self.nCPUs = 1
        self.rcon = 0

        self.pH = None
        self.ionicstr = None
        self.nlit = 100
        self.nonit = 5

        self.reduced_titration = True
        self.rt_cycles = 10
        self.rt_limit = 0.001

        self.pkai = False

        self.titrating_group = "Protein"

        self.pypka_input = "pypka_input.gro"

        self.sites = None
        self.ff_tautomers = {}

        self.pypka_optional_params = {}

        self.input_type = {
            "TimeStep": float,
            "InitTime": float,
            "temp": float,
            "EffectiveSteps": int,
            "RelaxSteps": int,
            "InitCycle": int,
            "EndCycle": int,
            "nCycles": int,
            "GROin": str,
            "TOPin": str,
            "DATin": str,
            "sysname": str,
            "titrating_group": str,
            "nCPUs": int,
            "pH": float,
            "ionicstr": float,
            "sites": list,
            "GroDIR": str,
            "NDXin": str,
            "rcon": float,
            "rt_cycles": int,
            "rt_limit": float,
            "reduced_titration": bool,
            "pkai": bool,
            "ffID": str,
        }

        # TODO implement '>=0' condition
        #                newEndCyle condition for InitCycle nCycles
        #                newEffectiveTime condition for EffectiveSteps TimeStep
        self.input_special_conditions = {}
        #    'TimeStep': '>=0',
        #    'InitTime': '>=0',
        #    'temp':     '>=0',
        #    'EffectiveSteps': '>=0'
        # }

    def get_simtime(self, cycle: int):
        simtime_begin = self.InitTime + self.EffectiveTime * cycle
        simtime_end = simtime_begin + self.EffectiveTime
        return simtime_begin, simtime_end

    def calcEndCyle(self):
        return self.InitCycle + self.nCycles

    def calcEffectiveTime(self):
        return self.EffectiveSteps * self.TimeStep

    def setFF(self):
        if self.ffID not in self.ff_options:
            raise IOError(
                "ffID parameter not valid. \nPlease choose one of available options: {}".format(
                    self.ffID, " ".join(self.ff_options.keys())
                )
            )

        self.ffDIR = "{0}/{1}.ff".format(self.script_dir, self.ffID)
        self.ff_dict = "{0}/protstates.dic".format(self.ffDIR)

        if "ffID" not in self.pypka_optional_params.keys():
            self.pypka_optional_params["ffID"] = self.ff_options[self.ffID]

        for line in read_ff_dict_section(self.ff_dict, "tautomers"):
            cols = line.split()
            if cols[0] not in self.ff_tautomers:
                self.ff_tautomers[cols[0]] = []
            self.ff_tautomers[cols[0]] += [(cols[1:])]

    def setFileNames(self):
        self.TOP = f"{self.sysname}.top"
        self.NDX = f"{self.sysname}.ndx"
        self.DAT = f"{self.sysname}.dat"

    def setPypKaParams(self):
        self.pypka_params = {
            "structure": self.pypka_input,
            "pH": str(self.pH),
            "epsin": 2,
            "ionicstr": self.ionicstr,
            "pbc_dimensions": 0,
            "ncpus": self.nCPUs,
            "convergence": 0.01,
            "clean_pdb": False,
            "CpHMD_mode": True,
            "sts": "sts_cphmd",
            "nlit": self.nlit,
            "nonit": self.nonit,
            "save_pdb": "delphi.pdb",
        }

        for param_name, param_value in self.pypka_optional_params.items():
            self.pypka_params[param_name] = param_value

        if self.pkai and self.reduced_titration:
            self.reduced_titration = False
            logging.warn("Reduced titration has been turned off")

    def check_mandatory_files(self):
        for f in (self.GROin, self.TOPin):
            if not os.path.isfile(f):
                raise IOError("Missing input file {}".format(f))

        for f in (self.NDXin, self.DATin):
            os.path.isfile(f)

        if not os.path.isfile("{}/gmx".format(self.GroDIR)):
            raise IOError("GROMACS executable not in dir {}".format(self.GroDIR))

    def read_input_mdp(self):
        def add_sites(mdp_sites):
            sites = []
            if mdp_sites == "all":
                return ["all"]

            for site in mdp_sites.split():
                if "N" in site or "C" in site:
                    sites.append(site)
                else:
                    try:
                        sites.append(int(site))
                    except:
                        raise IOError(
                            "sites has not been correctly defined. Example:\n"
                            "; sites = 1N 4 168 243C"
                        )
            return sites

        with open(self.MDPin) as f:
            for line in f:
                if line.startswith(";"):
                    line = line[1:]
                if line.startswith("SLURM"):
                    continue

                cleaned_line = remove_comments(line)

                parts = cleaned_line.split("=")
                param = parts[0].strip()

                if param in self:
                    param_value = parts[1].strip()
                    if param == "ref_t":
                        param_value = param_value.split()[0]

                    if param == "sites":
                        param_value = add_sites(param_value)

                    self[param] = param_value

                elif param.startswith("pypka_"):
                    param_name = param.strip("pypka_")
                    param_value = parts[1].strip()
                    self.pypka_optional_params[param_name] = param_value

        self.setFF()
        self.setFileNames()
        self.EndCycle = self.calcEndCyle()
        self.EffectiveTime = self.calcEffectiveTime()
        self.setPypKaParams()

        for i in self.__dict__:
            if self[i] == None:
                raise IOError("{} has not been defined.".format(i))

        self.check_mandatory_files()



import io
import logging
import os
import subprocess as sb
import time
from contextlib import redirect_stdout
from datetime import datetime, timedelta
from typing import Dict, Tuple, List

import sys

sys.path.insert(0, "/home/pedror/MMS@FCUL/pypka/")

from pypka import Titration

from pdbmender.formats import read_gro_line

from pypkamd.configs import Config
from pypkamd.misc import create_link, get_curtime, remove_comments


def create_cphdm_directory() -> None:
    os.system("rm -rf {}".format(Config.md_configs.tmpDIR))
    os.system("mkdir -p {}".format(Config.md_configs.tmpDIR))

    os.system(
        "cp -f {} {}/{}".format(
            Config.md_configs.TOPin, Config.md_configs.tmpDIR, Config.md_configs.TOP
        )
    )
    os.system(
        "cp -f {} {}/{}".format(
            Config.md_configs.GROin, Config.md_configs.tmpDIR, Config.md_configs.GRO
        )
    )
    os.system("cp -f {} {}/.".format(Config.md_configs.MDPin, Config.md_configs.tmpDIR))

    os.chdir("{}".format(Config.md_configs.tmpDIR))

    create_link("{}".format(Config.md_configs.ffDIR))
    # create_link("{}/residuetypes.dat".format(Config.md_configs.ffDIR))

    relax_mdp_new_params = [
        "nsteps",
        "Pcoupl",
        "freezegrps",
        "freezedim",
    ]

    with open(Config.md_configs.MDPin) as f, open(
        Config.md_configs.MDP, "w"
    ) as fmdp_new, open(Config.md_configs.MDP_relax, "w") as frelax_new:
        new_mdp_content = ""
        relax_mdp_content = ""
        cutoff_scheme = ""
        for line in f:
            cleaned_line = remove_comments(line)
            parts = cleaned_line.split("=")
            if len(parts) >= 2:
                param = parts[0].strip()
                if param not in relax_mdp_new_params:
                    relax_mdp_content += line
                if param == "cutoff-scheme":
                    cutoff_scheme = parts[1].split(";")[0].strip().lower()

            new_mdp_content += line

        preffix_energygrp_excl = ""
        if cutoff_scheme == "verlet":
            preffix_energygrp_excl = ";"

        relax_mdp_content += """
;Solvent relaxation parameters

nsteps = {0}
Pcoupl = No
;constraints = none
;energygrps = {1}
freezegrps = {1}
freezedim = Y Y Y
{2}energygrp_excl = {1} {1}
        """.format(
            Config.md_configs.RelaxSteps,
            Config.md_configs.titrating_group,
            preffix_energygrp_excl,
        )

        fmdp_new.write(new_mdp_content)
        frelax_new.write(relax_mdp_content)
    os.remove(Config.md_configs.MDPin)

    # Create a Index File
    if Config.md_configs.NDXin:
        os.system("cp ../{} {}".format(Config.md_configs.NDXin, Config.md_configs.NDX))
        print("cp ../{} {}".format(Config.md_configs.NDXin, Config.md_configs.NDX))
    else:
        new_ndx_cmd = "{}/gmx make_ndx -f {} -o {} " "-quiet".format(
            Config.md_configs.GroDIR, Config.md_configs.GRO, Config.md_configs.NDX
        )
        sb.run(
            new_ndx_cmd,
            shell=True,
            input=b"q\n",
            stdout=Config.md_configs.LOG,
            stderr=sb.STDOUT,
        )

    # Create a .tpr File to be used in the centering procedure
    new_tpr_cmd = (
        "{}/gmx grompp -f {} -c {} -p {} -n {} -o fixgro.tpr "
        "-maxwarn 1000 -quiet".format(
            Config.md_configs.GroDIR,
            Config.md_configs.MDP,
            Config.md_configs.GRO,
            Config.md_configs.TOP,
            Config.md_configs.NDX,
        )
    )
    sb.run(new_tpr_cmd, shell=True, stdout=Config.md_configs.LOG, stderr=sb.STDOUT)

    # Create a .dat file for fixbox
    if Config.md_configs.DATin:
        os.system("cp ../{} {}".format(Config.md_configs.DATin, Config.md_configs.DAT))
    else:
        with open(Config.md_configs.NDX) as f:
            save_trigger = False
            first_index = None
            last_index = None
            for line in f:
                if line.startswith("[ Protein ]"):
                    save_trigger = True
                elif line.startswith("["):
                    save_trigger = False
                elif save_trigger:
                    if not first_index:
                        first_index = line.split()[0]
                    last_index = line.split()[-1]

        with open(Config.md_configs.DAT, "w") as f:
            content = """# Molecular definitions file.
# All lines not starting with the following characters are ignored:
#   G : group name definition, followed by its molecule definitions.
#       a : molecule defined by atom (ordinal) index range.
#       n : molecule(s) defined by residue name.
#       g : molecules(s) defined through previously defined group.
#   A : group for assemblage stage, one line per (sequential) stage.
#   C : groups to be centered along the three box vectors (a,b,c or v1,v2,v3).

G Protein
#1st chain goes from 1st atom to atom 5743 (ordinal, not atom numbers)
a    {} {}

# Groups to be (sequentially) assembled:
A Protein

# Groups to be centered along each of the three box vectors:
C Protein Protein Protein W W W

# Use PBC (1) or not (0) along each of the three box vectors:
P Protein Protein Protein
    """.format(
                first_index, last_index
            )
            f.write(content)


def center_titrable_molecule(titrating_group: str, gro: str, ndx: str) -> str:
    # Centering procedure
    # effective.gro -> fixgro_input.gro -> centered.gro

    f_out_name = "centered.gro"

    fixgro_cmd = (
        "{}/gmx trjconv -f {} "
        "-o fixgro_input.gro -s fixgro.tpr -n {} "
        "-pbc mol -quiet; ".format(Config.md_configs.GroDIR, gro, ndx)
    )
    fixgro_cmd += "{}/fixbox fixgro_input.gro {} " "> {}".format(
        Config.md_configs.fixboxDIR, Config.md_configs.DAT, f_out_name
    )

    input_str = "{}\n".format(titrating_group)
    sb.run(
        fixgro_cmd,
        shell=True,
        input=str.encode(input_str),
        stdout=Config.md_configs.LOG,
        stderr=sb.STDOUT,
    )

    if os.path.isfile("fixgro_input.gro"):
        rm_cmd = "rm fixgro_input.gro"
        sb.run(rm_cmd, shell=True)
    else:
        raise Exception(
            f"Centering procedure failed. Check {Config.md_configs.LOG_fname} for more info."
        )

    return f_out_name


def run_dynamics(mdp: str, gro: str, top: str, ndx: str, sysname: str) -> None:
    tpr_cmd = (
        "{0}/gmx grompp -f {1} -c {2} -p {3} -n {4} -o {5}.tpr -po {5}_out.mdp "
        "-maxwarn 1000 -quiet".format(
            Config.md_configs.GroDIR, mdp, gro, top, ndx, sysname
        )
    )
    sb.run(tpr_cmd, shell=True, stdout=Config.md_configs.LOG, stderr=sb.STDOUT)

    mdrun_cmd = (
        "{0}/gmx mdrun -nt {1} -s {2}.tpr -x {2}.xtc -c {2}.gro -e {2}.edr "
        "-g {2}.log -o {2}.trr -rcon {3}".format(
            Config.md_configs.GroDIR,
            Config.md_configs.nCPUs,
            sysname,
            Config.md_configs.rcon,
        )
    )
    sb.run(mdrun_cmd, shell=True, stdout=Config.md_configs.LOG, stderr=sb.STDOUT)


def remove_freezed_velocities(top: str, relaxgro: str, effectivegro: str) -> None:
    freezed_atoms = top.index_atoms[Config.md_configs.titrating_group]

    final_gro = ""
    final_gro_footer = ""
    final_gro_atoms = {}

    # Get titrating atoms coords and velocities from EffectiveGro
    with open(effectivegro) as f_effective, open(relaxgro) as f_relax:
        line_counter = 0
        natoms_left = 0
        for line_eff, line_relax in zip(f_effective, f_relax):
            line_counter += 1
            if natoms_left > 0:
                natoms_left -= 1
                (aname, anumb, resname, resnumb, x, y, z) = read_gro_line(line_eff)
                if anumb in freezed_atoms:
                    final_gro_atoms[anumb] = line_eff
                else:
                    final_gro_atoms[anumb] = line_relax
            elif line_counter == 2:
                natoms_left = int(line_eff.strip())
                final_gro += line_eff
            elif line_counter == 1:
                final_gro += line_eff
            else:
                final_gro_footer += line_eff

    for atom in sorted(final_gro_atoms.keys()):
        final_gro += final_gro_atoms[atom]
    final_gro += final_gro_footer

    with open(relaxgro, "w") as f_new:
        f_new.write(final_gro)


def concat_results(
    runname: str,
    cycle: int,
    effective_name: str,
    init_time: float,
    simtime_begin: float,
) -> None:
    if cycle == 0:
        edr_in = "{}.edr".format(effective_name)
        edr_out = "{}.edr".format(runname)
        edr_stdin = "{}\nc".format(simtime_begin)

        xtc_in = "{}.xtc".format(effective_name)
        xtc_out = "{}.xtc".format(runname)
        xtc_cmd = "{}/gmx trjconv -f {} -o {} -t0 {} -s {}.tpr".format(
            Config.md_configs.GroDIR, xtc_in, xtc_out, init_time, effective_name
        )
        xtc_stdin = "System"
    else:
        edr_cur = "{}.edr".format(runname)
        edr_in = "{} {}.edr".format(edr_cur, effective_name)
        edr_out = "aux.edr"
        edr_stdin = "{}\n{}".format(init_time, simtime_begin)

        xtc_cur = "{}.xtc".format(runname)
        xtc_in = "{} {}.xtc".format(xtc_cur, effective_name)
        xtc_out = "aux.xtc"
        xtc_cmd = "{}/gmx trjcat -f {} -o {} -settime".format(
            Config.md_configs.GroDIR, xtc_in, xtc_out
        )
        xtc_stdin = "{}\nc\n".format(init_time)

    edr_cmd = "{}/gmx eneconv -o {} -f {} -settime".format(
        Config.md_configs.GroDIR, edr_out, edr_in
    )
    sb.run(
        edr_cmd,
        shell=True,
        input=str.encode(edr_stdin),
        stdout=Config.md_configs.LOG,
        stderr=sb.STDOUT,
    )

    sb.run(
        xtc_cmd,
        shell=True,
        input=str.encode(xtc_stdin),
        stdout=Config.md_configs.LOG,
        stderr=sb.STDOUT,
    )

    if cycle != 0:
        os.rename(edr_out, edr_cur)
        os.rename(xtc_out, xtc_cur)

    save_logs_cmd = "cat {}.log >> {}.log".format(effective_name, runname)
    sb.run(save_logs_cmd, shell=True)


def clean_dir(effective_name: str, relax_name: str, pypka_input: str) -> None:
    clean_cmd = (
        "rm -f {0}.edr {0}.log {0}.xtc {0}.tpr {1}.edr"
        " {1}.gro {1}.log {1}.tpr {1}.xtc state.cpt"
        " state_prev.cpt mdout.mdp {2} *#".format(
            effective_name, relax_name, pypka_input
        )
    )
    sb.run(clean_cmd, shell=True)


def final_cleanup(sysname: str, tmpDIR: str, effective_name: str) -> None:
    os.system("cp {0}.gro ../{1}.gro".format(effective_name, sysname))
    os.system("cp {0}.edr {0}.log {0}.mocc {0}.occ {0}.xtc {0}.top ../".format(sysname))
    os.chdir("../")
    os.system("rm -r {}".format(tmpDIR))


def write_pypka_inputgro(f_centered_name: str, nontitrating_res: str) -> None:
    new_lines = []
    with open(f_centered_name) as f:
        line_counter = 0
        natoms_left = 0
        for line in f:
            line_counter += 1
            if natoms_left > 0:
                natoms_left -= 1
                resnumb = read_gro_line(line)[3]

                if resnumb in nontitrating_res:
                    resname = nontitrating_res[resnumb]
                    line = "{}{}{}".format(line[:5], resname, line[8:])

            elif line_counter == 2:
                natoms_left = int(line.strip())

            new_lines.append(line)

    with open(Config.md_configs.pypka_input, "w") as f_new:
        f_new.write("".join(new_lines))


def run_pbmc(
    params: Dict, sites: List, pH: float, offset: int, fixed_sites: Dict
) -> Tuple[Dict[int, int], Dict[int, float], Dict[int, list]]:
    prot_states = {}
    prot_avgs = {}
    taut_probs = {}
    if len(sites) > 0:
        with redirect_stdout(io.StringIO()) as f:
            tit = Titration(params, sites={"A": sites}, fixed_sites={"A": fixed_sites})

        Config.md_configs.LOG.write(f.getvalue())
        print("\r" + " " * 90, end="\r")

        for site in tit:
            resnumb = site.getResNumber()
            resname = site.getName()
            if resname in ("NTR", "CTR"):
                resnumb += offset
            prot_states[resnumb] = site.getFinalState(pH)
            prot_avgs[resnumb] = site.getTitrationCurve()[pH]
            taut_probs[resnumb] = site.getTautsProb(pH)

    return (prot_states, prot_avgs, taut_probs)


def run_pkai(
    fname, sites, pH, offset
) -> Tuple[Dict[int, int], Dict[int, float], Dict[int, list]]:
    try:
        Pege
    except NameError:
        from pege import Pege

    prot_states = {}
    prot_avgs = {}
    taut_probs = {}

    protein = Pege(fname, save_final_pdb=False, fix_pdb=False)

    for site in sites:
        resnumb = site
        if isinstance(resnumb, str):
            resnumb = int(site[:-1]) + offset

        (
            prot_states[resnumb],
            prot_avgs[resnumb],
            taut_probs[resnumb],
        ) = protein.get_residue_taut_probs("A", site, pH)

    return (prot_states, prot_avgs, taut_probs)


def run_cphmd(top: str) -> None:
    cur_time, prev_time = None, None
    cycle_times = []
    for cycle in range(Config.md_configs.InitCycle, Config.md_configs.EndCycle):
        cur_time = time.time()

        simtime_begin, simtime_end = Config.md_configs.get_simtime(cycle)

        info = "{} | Starting CpHMD Cycle #{}".format(get_curtime(), cycle)
        logging.info(info)

        if prev_time:
            cycle_time = cur_time - prev_time
            cycle_times.append(cycle_time)
            cycles_left = Config.md_configs.EndCycle - cycle
            avg_time = sum(cycle_times) / len(cycle_times) * cycles_left
            end_guess = datetime.now() + timedelta(seconds=avg_time)
            info = "{} | Estimated Time of Completion: {}".format(
                get_curtime(), end_guess.strftime("%d %B %Y %H:%M")
            )
            logging.info(info)

        ### PB/MC ###
        f_centered_name = center_titrable_molecule(
            Config.md_configs.titrating_group,
            Config.md_configs.GRO,
            Config.md_configs.NDX,
        )

        if not Config.md_configs.pkai:
            info = "{} | PB/MC - titrating {} sites".format(
                get_curtime(), len(top.titrating_sites)
            )
            if top.fixed_sites:
                info += " and {} fixed states".format(len(top.fixed_sites))
            logging.info(info)

            write_pypka_inputgro(f_centered_name, top.nontit_tautomers)
            os.system(f"cp {Config.md_configs.pypka_input} pypka_input_{cycle}.gro")

            pb_prot_states, pb_prot_avgs, pb_taut_probs = run_pbmc(
                Config.md_configs.pypka_params,
                top.titrating_sites[:],
                Config.md_configs.pH,
                top.offset,
                {k: v["state"] for k, v in top.fixed_sites.items()},
            )
        else:
            info = "{} | AI/MC - titrating {} sites".format(
                get_curtime(), len(top.titrating_sites)
            )
            logging.info(info)

            pb_prot_states, pb_prot_avgs, pb_taut_probs = run_pkai(
                f_centered_name,
                top.titrating_sites[:],
                Config.md_configs.pH,
                top.offset,
            )

        top.update(pb_prot_states, pb_prot_avgs, pb_taut_probs)
        top.write_top_file()

        ### MD ###
        info = "{} | MD    - solvent relaxation for {} steps".format(
            get_curtime(), Config.md_configs.RelaxSteps
        )
        logging.info(info)
        os.system(f"cp {Config.md_configs.TOP} topol_{cycle}.top")

        run_dynamics(
            Config.md_configs.MDP_relax,
            Config.md_configs.GRO,
            Config.md_configs.TOP,
            Config.md_configs.NDX,
            Config.md_configs.relax_name,
        )

        remove_freezed_velocities(
            top, Config.md_configs.GRO_relax, Config.md_configs.GRO
        )

        info = "{} | MD    - production of {} steps".format(
            get_curtime(), Config.md_configs.EffectiveSteps
        )
        logging.info(info)

        run_dynamics(
            Config.md_configs.MDP,
            Config.md_configs.GRO_relax,
            Config.md_configs.TOP,
            Config.md_configs.NDX,
            Config.md_configs.effective_name,
        )

        ### Concatenate Results ###
        concat_results(
            Config.md_configs.sysname,
            cycle,
            Config.md_configs.effective_name,
            Config.md_configs.InitTime,
            simtime_begin,
        )
        clean_dir(
            Config.md_configs.effective_name,
            Config.md_configs.relax_name,
            Config.md_configs.pypka_input,
        )

        prev_time = cur_time



import os
from datetime import datetime
from typing import List

from pypka import getTitrableSites
from pdbmender.formats import gro2pdb, read_pdb_line


def get_curtime() -> str:
    return datetime.today().strftime("%Y-%m-%d %H:%M")


def remove_comments(line: str, comment_delimiter: str = ";") -> str:
    if len(comment_delimiter) == 1:
        return line.split(comment_delimiter)[0]
    return remove_comments(line.split(comment_delimiter[1:])[0])


def create_link(path: str) -> None:
    os.system("ln -s -f {} .".format(path))


def get_titrable_sites(groname: str) -> List[int]:
    gro2pdb(groname, "tmp.pdb")
    sites, _ = getTitrableSites("tmp.pdb", ser_thr_titration=False)

    potential = []
    confirmed = []
    with open("tmp.pdb") as f:
        for line in f:
            if line.startswith("ATOM "):
                atom, _, res, chain, resnumb, *_ = read_pdb_line(line)
                if resnumb in sites["A"] and res == "CYS":
                    if resnumb not in potential:
                        potential.append(resnumb)
                    if atom == "HG" and resnumb not in confirmed:
                        confirmed.append(resnumb)
    to_del = set(potential) - set(confirmed)
    sites = set(sites["A"]) - to_del

    return list(sites)


def check_convert_termini(site: str, offset: int) -> int:
    if isinstance(site, str) and site[-1] in "NC":
        site = offset + int(site[:-1])
    return site

def read_ff_dict_section(fname, section):
    trigger_read = False
    with open(fname) as f_dict:
        for line in f_dict:
            if "[" in line and "]" in line:
                if section in line:
                    trigger_read = True
                else:
                    trigger_read = False
            elif trigger_read:
                yield line


import logging
import sys
import pathlib

sys.path.insert(0, str(pathlib.Path(__file__).parent.parent))

from pypkamd.cli import CLI

logging.basicConfig(format="%(message)s", level=logging.INFO)


def main():
    CLI()


if __name__ == "__main__":
    main()



import argparse
import logging
import time

from pypkamd.configs import Config
from pypkamd.cphmd import create_cphdm_directory, final_cleanup, run_cphmd
from pypkamd.misc import get_curtime
from pypkamd.topology import Topology
from pypkamd import __version__
from pypka import __version__ as pypka_version
from pdbmender import __version__ as pdbmender_version


def CLI():
    init_time = time.time()
    parser = argparse.ArgumentParser(description="PypKa-MD")
    parser.add_argument(
        "mdp", type=str, help="MDP file", action="store", default="system.mdp"
    )
    parser.add_argument("-v", "--version", action="version", version=__version__)

    args = parser.parse_args()

    logging.info("### Starting PypKa-MD ###")

    Config.storeParams(args.mdp)

    if Config.md_configs.pkai:
        from pege import __version__ as pege_version

        pege_info = f"PEGE version: {pege_version}\n"

    info = (
        f"Simulation run by {Config.md_configs.USER} @ {Config.md_configs.HOST}\n"
        f"PypKa-MD version: {__version__}\nPypKa version: {pypka_version}\n"
        f"pdbmender version: {pdbmender_version}\n{pege_info if Config.md_configs.pkai else ''}"
        f"Initial time: {get_curtime()}"
    )
    logging.info(info)

    top = Topology(Config.md_configs)

    create_cphdm_directory()
    top.read_index(Config.md_configs.NDX)

    run_cphmd(top)

    final_cleanup(
        Config.md_configs.sysname,
        Config.md_configs.tmpDIR,
        Config.md_configs.effective_name,
    )

    end_time = time.time()
    duration = end_time - init_time
    duration_formated = time.strftime("%H h %M m %S s", time.gmtime(duration))
    logging.info("\nCpHMD block duration: {}".format(duration_formated))
    logging.info("### PypKa-MD finished succesfully ###")



import logging
import random
from typing import Optional, List, Tuple, Dict

from pypkamd.misc import (
    check_convert_termini,
    get_titrable_sites,
    remove_comments,
    read_ff_dict_section,
)
from pypkamd.constants import TITRABLE_AAS
from pypkamd.configs import Config


class Topology:
    def __init__(self, configs: Config):

        self.sections = ["atoms", "bonds", "pairs", "angles", "dihedrals", "impropers"]
        self.section = {section: [] for section in self.sections}
        self.section["others"] = {}

        self.tit_atoms = {}
        self.tit_atypes = {}
        self.tit_charges = {}
        self.tit_bonds = {}
        self.tit_angles = {}
        self.tit_dihedrals = {}
        self.tit_impropers = {}

        self.prop_types = {
            "t": (1, self.tit_atypes),
            "q": (1, self.tit_charges),
            "b": (2, self.tit_bonds, self.section["bonds"]),
            "a": (3, self.tit_angles, self.section["angles"]),
            "d": (4, self.tit_dihedrals, self.section["dihedrals"]),
            "i": (4, self.tit_impropers, self.section["impropers"]),
        }

        self.offset = None
        self.ters = {}

        self.config_vars = configs

        if configs.sites == ["all"]:
            configs.sites = get_titrable_sites(configs.GROin)

            logging.info(
                " {} titrable sites have been identified".format(len(configs.sites))
            )

        self.all_sites = configs.sites[:]
        self.titrating_sites = configs.sites[:]
        self.rt_cur_cycle = 0
        self.fixed_sites = {}

        self.read_input_top(
            configs.TOPin, configs.sites, configs.ffDIR.split("/")[-1], configs.ff_dict
        )
        self.read_ff_dict(configs.ff_dict)

        self.nontit_tautomers = {}
        self.get_non_tit_tauts(configs.ff_tautomers)

        self.index_atoms = {configs.titrating_group: []}

        self.titres_states = {}
        self.mocc = {}
        self.occ = {}
        for resnumb in configs.sites:
            if type(resnumb) == str:
                resnumb = self.offset + int(resnumb[:-1])

            self.titres_states[resnumb] = []
            self.mocc[resnumb] = []
            self.occ[resnumb] = []

        self.f_output = configs.TOP
        self.f_occ = configs.sysname + ".occ"
        self.f_mocc = configs.sysname + ".mocc"

    def get_nres(self) -> int:
        return int(self.section["atoms"][-1][2])

    def get_atom(self, anumb: str) -> Optional[int]:
        for atom in self.section["atoms"]:
            if atom[0] == anumb:
                return atom
        return None

    def get_titrable_resnames(self) -> List[str]:
        resnames = []
        for resnumb, atom_i_list in self.tit_atoms.items():
            if resnumb < self.offset:
                atom_i = atom_i_list[0]
                atom = self.section["atoms"][atom_i]
                resname = atom[3]
                resnames.append(resname)
            else:
                ter_numb = resnumb - self.offset
                resname = self.ters[ter_numb]
                resnames.append(resname)

        return list(set(resnames))

    def atom_is_titrable(self, anumb: int) -> Optional[int]:
        anumb = int(anumb) - 1
        for residue, atoms in self.tit_atoms.items():
            if anumb in atoms:
                return residue
        return False

    def get_atoms(self, resname: str, aname: str) -> Tuple[int, str, str]:
        # print("###################################", resname, aname)
        for atom in self.section["atoms"]:
            anumb, resnumb, resname_, aname_ = atom[0], int(atom[2]), atom[3], atom[4]
            tit_residue = self.atom_is_titrable(anumb)
            if (
                resnumb in self.ters.keys()
                and int(anumb) in self.tit_atoms[resnumb + self.offset]
            ):
                resname_ = self.ters[resnumb]
            if aname == aname_ and tit_residue and resname == resname_:
                yield (anumb, resname, aname)

    def search_by_anames(
        self, search_item: str, resname: str, anames: str
    ) -> Tuple[int, list]:
        anumbs = []
        for aname in anames:
            for i in self.get_atoms(resname, aname):
                anumbs.append(i[0])

        for i, item in enumerate(search_item):
            item_anames = []

            for anumb in item[: len(anames)]:
                atom = self.get_atom(anumb)
                anumb, resname_, aname_ = atom[0], atom[3], atom[4]
                if anumb in anumbs:
                    item_anames.append(aname_)

            if item_anames == anames:
                yield (i, item)

    def get_resnumb_from_anumb(self, anumb: str) -> int:
        for atom in self.section["atoms"]:
            anumb_ = atom[0]
            if anumb_ == anumb:
                return atom[2]

    def read_input_top(
        self, f_top: str, titrating_resnumbs: List[int], ffDIR: str, ff_dict: str
    ):
        def save_other_lines(last_section: str, line: str):
            if last_section == None:
                last_section = "begin"
            elif last_section == False:
                last_section = "end"
            if last_section not in self.section["others"]:
                self.section["others"][last_section] = []
            self.section["others"][last_section].append(line)

        ter_resnumbs = {}
        tit_resnumbs = []
        for i in titrating_resnumbs:
            if type(i) == str:
                numb = int(i.strip("NC"))
                ter_resnumbs[numb] = i[-1]
                self.ters[numb] = "NTR"
            else:
                tit_resnumbs.append(i)

        cph_ready_resnames = {i: ii for i, ii in TITRABLE_AAS.items()}

        termini_atoms = {}
        for line in read_ff_dict_section(ff_dict, "replace"):
            cols = line.split()
            resname = cols[0]
            atom_type = cols[2]

            if resname[:2] not in ("NT", "CT"):
                continue

            if resname not in termini_atoms:
                termini_atoms[resname] = []
            if atom_type not in termini_atoms[resname]:
                termini_atoms[resname].append(atom_type)

        last_section = None
        section = "begin"
        tmp_ter_atoms = {}
        with open(f_top) as f:
            for line in f:
                clean_line = remove_comments(line, comment_delimiter=";").strip()
                if not clean_line:
                    if not line.strip() and last_section != None:
                        last_section, section = section, False
                    continue

                if "#include" in line and "./" in line:
                    ff = line.split("./")[1].split("/")[0]
                    assert (
                        ff == ffDIR
                    ), "Topology file error\nLine {}\nShould be replaced by: {}".format(
                        line, line.replace(ff, ffDIR)
                    )

                if clean_line.startswith("[ "):
                    tmp_section = clean_line.strip("[] \n")
                    if self.section["dihedrals"] and "dihedrals" == tmp_section:
                        tmp_section = "impropers"

                    if tmp_section in self.sections:
                        last_section, section = section, tmp_section

                    if section not in self.sections:
                        save_other_lines(last_section, clean_line)

                elif section in self.sections:
                    cols = clean_line.split()
                    self.section[section].append(cols)

                    resnumb = int(cols[2])
                    if section == "atoms":
                        resname = cols[3]
                        atype = cols[4]
                        atom_i = len(self.section["atoms"]) - 1
                        add_ter_atom = False

                        if resnumb in tit_resnumbs:
                            if resname[:2] in cph_ready_resnames.keys():
                                self.section["atoms"][-1][3] = cph_ready_resnames[
                                    resname[:2]
                                ]

                        if resnumb in ter_resnumbs:
                            ter_type = ter_resnumbs[resnumb]
                            if ter_type == "N":
                                if (
                                    resname not in ("GLY", "PRO")
                                    and atype in termini_atoms["NT"]
                                ):
                                    self.ters[resnumb] = "NT"
                                    add_ter_atom = True
                                    # self.section["atoms"][-1][3] = "NT"

                                elif (
                                    resname in ("GLY", "PRO")
                                    and atype
                                    in termini_atoms["NTGLY"] + termini_atoms["NTPRO"]
                                ):
                                    self.ters[resnumb] = "NT" + resname
                                    add_ter_atom = True
                            elif ter_type == "C":
                                if atype in termini_atoms["CT"]:
                                    self.ters[resnumb] = "CT"
                                    add_ter_atom = True
                                    # self.section["atoms"][-1][3] = "CT"

                            if add_ter_atom:
                                if resnumb not in tmp_ter_atoms:
                                    tmp_ter_atoms[resnumb] = []
                                tmp_ter_atoms[resnumb].append(atom_i)

                        if (
                            resnumb in tit_resnumbs
                            and atype not in ("N", "H", "CA", "C", "O")
                            and not add_ter_atom
                        ):
                            if resnumb not in self.tit_atoms:
                                self.tit_atoms[resnumb] = []
                            self.tit_atoms[resnumb].append(atom_i)

                else:
                    save_other_lines(last_section, clean_line)

        self.offset = 10 ** (len(str(self.get_nres())) + 1)

        for resnumb, atoms in tmp_ter_atoms.items():
            resnumb += self.offset
            self.tit_atoms[resnumb] = atoms

        # from pprint import pprint

        # pprint(self.tit_atoms)
        # pprint(self.tit_angles)
        # pprint(self.section["atoms"])
        # pprint(self.ters)
        # exit()

    def read_ff_dict(self, ff_dict: dict) -> None:

        tit_res_cph = self.get_titrable_resnames()

        for line in read_ff_dict_section(ff_dict, "replace"):
            cols = line.split()
            resname = cols[0]
            line_type = cols[1]
            natoms = self.prop_types[line_type][0]
            n_tauts = int(cols[2 + natoms])
            assert n_tauts + natoms + 3 == len(
                cols
            ), "There is an inconsistency in the ff_dict file"

            if resname not in tit_res_cph:
                # print(resname, tit_res)
                continue

            properties = cols[-n_tauts:]
            prop_obj = self.prop_types[line_type][1]

            if line_type in ("t", "q"):
                aname = cols[2]
                for atom in self.get_atoms(resname, aname):
                    prop_obj[atom[0]] = properties

            elif line_type in ("b", "a", "d", "i"):
                anames = cols[2 : 2 + natoms]
                search_in = self.prop_types[line_type][2]

                for i, item in self.search_by_anames(search_in, resname, anames):
                    resnumb = self.get_resnumb_from_anumb(item[0])
                    prop_obj[i] = (resnumb, item[:natoms], properties)

        # from pprint import pprint
        # pprint(self.tit_angles)
        # pprint(self.prop_types["a"])
        # pprint(self.tit_charges)
        # exit()

    def get_non_tit_tauts(self, titratable_residues) -> None:
        anames = {
            resname: [i[1] for i in cols]
            for resname, cols in titratable_residues.items()
        }
        for atom in self.section["atoms"]:
            _, _, resnumb, resname, aname, _, charge, _ = atom

            if (
                resname in titratable_residues
                and aname in anames[resname]
                and int(resnumb) not in self.all_sites
            ):
                taut_numb = None
                for taut in titratable_residues[resname]:
                    taut_numb, taut_aname, taut_acharge = taut
                    if taut_aname == aname and taut_acharge == charge:
                        break
                self.nontit_tautomers[int(resnumb)] = taut_numb

    def read_index(self, ndx: str) -> None:
        index_groups = self.index_atoms.keys()
        with open(ndx) as f:
            save_group = False
            for line in f:
                if "[" in line and "]" in line:
                    index_group = line.strip("[ ]\n")
                    if index_group in index_groups:
                        save_group = True
                    else:
                        save_group = False
                elif save_group:
                    try:
                        indexes = [int(i) for i in line.strip().split()]
                        self.index_atoms[index_group] += indexes
                    except:
                        msg = "Index group {} in file {} " "is not valid.".format(
                            index_group, ndx
                        )
                        raise IOError(msg)

    def sample_fixed(self, prob_states: Dict[int, int]) -> Dict[int, int]:
        for site, fixed_details in self.fixed_sites.items():
            tauts = list(range(len(fixed_details["taut_probs"])))
            new_state = random.choices(tauts, fixed_details["taut_probs"])[0]

            if isinstance(site, str) and site[-1] in "NC":
                s = int(site[:-1]) + self.offset
            else:
                s = site

            prob_states[s] = new_state
            self.fixed_sites[site]["state"] = new_state
            self.mocc[s].append(fixed_details["prob"])

        return prob_states

    def update(self, new_states, new_protavg, new_tautprobs):
        if self.config_vars["reduced_titration"]:
            self.rt_cur_cycle += 1

            if self.fixed_sites:
                new_states = self.sample_fixed(new_states)

            if self.config_vars["rt_cycles"] == self.rt_cur_cycle:
                self.rt_cur_cycle = 0
                self.titrating_sites = self.config_vars["sites"][:]
                self.fixed_sites = {}

        # Save the current average protonation
        for site, avg in new_protavg.items():
            self.mocc[site].append(avg)

            if self.config_vars["reduced_titration"] and self.rt_cur_cycle == 1:
                if (
                    avg < self.config_vars["rt_limit"]
                    or avg > 1 - self.config_vars["rt_limit"]
                ):
                    if site > self.offset:
                        s = site - self.offset
                        if "NT" in self.ters[s]:
                            s = "{}N".format(s)
                        elif "CT" in self.ters[s]:
                            s = "{}C".format(s)
                        else:
                            raise Exception("Something is wrong!")
                    else:
                        s = site

                    site_i = self.titrating_sites.index(s)
                    del self.titrating_sites[site_i]

                    self.fixed_sites[s] = {
                        "prob": avg,
                        "state": new_states[site],
                        "taut_probs": new_tautprobs[site],
                    }

        # print(self.titres_states[site])
        # print("rt cycle:", self.rt_cur_cycle)
        # print("fixed:", self.fixed_sites)
        # print("titrating next:", self.titrating_sites)

        # Deal with the new protonation states
        for site, state in new_states.items():
            self.titres_states[site] = state
            self.occ[site].append(state)

            for atom_id in self.tit_atoms[site]:
                atom = self.section["atoms"][atom_id]
                anumb = atom[0]

                if anumb in self.tit_atypes:
                    atom[1] = self.tit_atypes[anumb][state]

                if anumb in self.tit_charges:
                    atom[6] = self.tit_charges[anumb][state]

        for prop_code, prop_objs in self.prop_types.items():
            if prop_code in ("q", "t"):
                continue
            natoms, titprop_list, top_list = prop_objs
            for index, item in titprop_list.items():
                resnumb, anumbs = int(item[0]), item[1]

                assert top_list[index][:natoms] == anumbs

                if resnumb in self.ters:
                    resnumb += self.offset

                new_state = self.titres_states[resnumb]
                new_prop = item[2][new_state]

                top_list[index][-1] = new_prop

        self.write_prot_file(self.occ, self.f_occ, "occ")
        self.write_prot_file(self.mocc, self.f_mocc, "mocc")

    def get_ordered_sites(self):
        def get_resnumber(res):
            if isinstance(res, str):
                res = int(res[:-1])
            return res

        sorted_sites = []
        termini = []
        for res in self.all_sites:
            if isinstance(res, str):
                termini.append(res)
            else:
                sorted_sites.append(res)
        sorted_sites.sort()
        for res in termini:
            res_numb = int(res[:-1])
            if not sorted_sites:
                sorted_sites.append(res)
                continue
            sorted_res = get_resnumber(sorted_sites[0])

            i = 0
            while sorted_res < res_numb and i + 2 <= len(sorted_sites):
                i += 1
                sorted_res = get_resnumber(sorted_sites[i])
                # print(sorted_res, res_numb, i, len(sorted_sites))

            if res_numb > sorted_res:
                i += 1

            if res[-1] == "C" and sorted_res <= res_numb:
                i += 1

            sorted_sites.insert(i, res)
        return sorted_sites

    def write_prot_file(self, prot_obj, fprot_name, mode):
        content = "# "
        nframes = len(list(prot_obj.values())[0])

        for site in self.get_ordered_sites():
            if mode == "mocc":
                content += "{:>8} ".format(site)
            elif mode == "occ":
                content += "{:>4} ".format(site)

        content += "\n  "
        for i in range(nframes):
            for site in self.get_ordered_sites():
                site = check_convert_termini(site, self.offset)
                site_occ = prot_obj[site][i]
                if mode == "mocc":
                    content += "{:8.6f} ".format(site_occ)
                elif mode == "occ":
                    content += "{:4} ".format(site_occ)
            content += "\n  "
        with open(fprot_name, "w") as f_new:
            f_new.write(content)

    def write_top_file(self):
        def seclist_to_str(seclist, extra_spaces=5):
            sec_str = ""
            maxsize = len(self.section["atoms"][-1][0]) + extra_spaces
            for line in seclist:
                tmp_line = ""
                for item in line[:-1]:
                    col = "{}".format(item)
                    while len(col) < maxsize:
                        col = " " + col
                    tmp_line += col
                tmp_line += " " * maxsize
                tmp_line += line[-1]
                sec_str += tmp_line + "\n"
            return sec_str

        top_content = ""
        if "begin" in self.section["others"]:
            top_content += "\n".join(self.section["others"]["begin"]) + "\n"

        top_content += "[ atoms ]\n"
        top_content += seclist_to_str(self.section["atoms"], extra_spaces=8) + "\n"

        # print(seclist_to_str(self.section["atoms"], extra_spaces=8))

        for section in self.sections[1:]:
            section_name = section
            if section == "impropers":
                section_name = "dihedrals"

            top_content += "[ {} ]\n".format(section_name)
            top_content += seclist_to_str(self.section[section])
            if section in self.section["others"]:
                top_content += "\n".join(self.section["others"][section]) + "\n"

        if "end" in self.section["others"]:
            top_content += "\n".join(self.section["others"]["end"]) + "\n"

        final_content = ""
        for line in top_content.splitlines():
            if line.startswith("[ ") or line.startswith("#ifdef"):
                line = "\n" + line
            final_content += line + "\n"

        with open(self.f_output, "w") as ftop_new:
            ftop_new.write(final_content)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pypkamd']

package_data = \
{'': ['*'], 'pypkamd': ['CHARMM36mpH.ff/*', 'G54A7pH.ff/*', 'fixbox/*']}

install_requires = \
['pypka==2.10.0']

entry_points = \
{'console_scripts': ['pypkamd = pypkamd.__main__:main']}

setup_kwargs = {
    'name': 'pypkamd',
    'version': '0.5.0',
    'description': 'PypKa + GROMACS = an awesome Constant-pH Molecular Dynamics implementation',
    'long_description': '# PypKa-MD\n\nPypKa + MD = constant-pH molecular dynamics\n\nImplementation of the stochastic titration method <sup>1</sup>\n\n[1] Baptista *et al.*, J. Chem. Phys. 117, 4184 (2002) DOI: 10.1063/1.1497164\n\n## Installation\n\n```\npython3 -m pip install pypkamd\n```\n\n## Dependencies\n\nBoth PypKa and GROMACS are required to be installed in the system.\n\n- PypKa >= 2.7.1\n- GROMACS >=5.1.5\n\nWhen running in pKAI-MD mode there are extra dependencies:\n\n- pege >= 1.1.1\n- torch_geometric >= 2.0.0\n\nPlease refer to the installation guide of [torch geometric](https://pytorch-geometric.readthedocs.io/en/latest/notes/installation.html) to install the proper version in accordance to your CUDA and OS.\n\n```\npython3 -m pip install pege\n# EXAMPLE FOR CUDA10.2\n# python3 -m pip install torch-scatter torch-sparse torch-cluster torch-spline-conv torch-geometric -f https://data.pyg.org/whl/torch-1.10.0+cu102.html\n```\n\n\n## Usage\n\nUpon installation a PypKa-MD executable should have been added to your bin. You may call it directly giving as an argument a modified GROMACS .mdp input file to include Constant-pH specific variables.\n\n```\npypkamd System.mdp\n```\n\nIn case the executable as not been added to your bin, you may use:\n\n```\npython3 -m pypkamd System.mdp\n```\n\nYou may find an example .mdp file in /utils/cphmd.mdp. \n\n```\n; GROin = system_000.gro     ; input GRO file\n; TOPin = system_000.top     ; input TOP file\n; DATin = fixgro.dat         ; input DAT file (to be removed)\n; NDXin = system.ndx         ; input NDX file\n; sysname = system_001       ; output files root name\n; sites = all                ; to be titrating sites in the protein\n; titrating_group = Protein  ; index group of the protein\n; nCycles = 50               ; number of CpHMD cycles\n                            ;; total simulation time = nCycles * tau_prot\n                            ;; 1ns = 50 * 20ps\n; nCPUs = 4                  ; number of CPUs to be used\n; pH = 7.0                   ; pH value of the protonation states sampling\n; ionicstr = 0.1             ; ionic strength used in PB\n; GroDIR="/gromacs/gromacs-5.1.5_pH_I/bin/" ; GROMACS bin path\n```',
    'author': 'Pedro Reis',
    'author_email': 'pdreis@fc.ul.pt',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/mms-fcul/PypKa-MD',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6.2,<4.0',
}


setup(**setup_kwargs)
